# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime

from minio.compat import urlencode
from minio.error import ResponseError
from minio.credentials import Credentials
from minio.helpers import get_sha256_hexdigest
from minio.signer import sign_v4

from .credentials import Expiry, Provider
from .parsers import parse_assume_role


class AssumeRoleProvider(Provider):
    region = 'us-east-1'

    # AWS STS support GET and POST requests for all actions. That is, the API does not require you to
    # use GET for some actions and POST for others. However, GET requests are subject to the limitation
    # size of a URL; although this limit is browser dependent, a typical limit is 2048 bytes. Therefore,
    # for Query API requests that require larger sizes, you must use a POST request.
    method = 'POST'

    def __init__(self, mc, RoleArn=None, RoleSessionName=None, Policy=None, DurationSeconds=None):
        self._minio_client = mc
        self._expiry = Expiry()
        self._DurationSeconds = DurationSeconds
        self._RoleArn = "arn:xxx:xxx:xxx:xxxx" if RoleArn is None else RoleArn
        self._RoleSessionName = "anything" if RoleSessionName is None else RoleSessionName
        self._Policy = Policy

        super(Provider, self).__init__()

    def retrieve(self):

        query = {
            "Action": "AssumeRole",
            "Version": "2011-06-15",
            "RoleArn": self._RoleArn,
            "RoleSessionName": self._RoleSessionName,
        }

        # Add optional elements to the request
        if self._Policy is not None:
            query["Policy"] = self._Policy

        if self._DurationSeconds is not None:
            query["DurationSeconds"] = str(self._DurationSeconds)

        url = self._minio_client._endpoint_url + "/"
        content = urlencode(query)
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
            'User-Agent': self._minio_client._user_agent
        }

        # Create signature headers
        content_sha256_hex = get_sha256_hexdigest(content)
        signed_headers = sign_v4(self.method, url, self.region, headers,
                                 self._minio_client._credentials,
                                 content_sha256=content_sha256_hex,
                                 request_datetime=datetime.utcnow(),
                                 service_name='sts'
                                 )
        response = self._minio_client._http.urlopen(self.method, url,
                                                    body=content,
                                                    headers=signed_headers,
                                                    preload_content=True)

        if response.status != 200:
            raise ResponseError(response, self.method).get_exception()

        # Parse the XML Response - getting the credentials as a Values instance.
        credentials_value, expiry = parse_assume_role(response.data)
        self._expiry.set_expiration(expiry)

        return credentials_value

    def is_expired(self):
        return self._expiry.is_expired()



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from unittest import TestCase
from minio.credentials.file_minio_client import FileMinioClient
from minio.credentials.credentials import Value

from nose.tools import eq_


class FileMinioClientTest(TestCase):

    def test_file_minio_(self):
        # clear environment
        os.environ.clear()
        # get provider
        provider = FileMinioClient('minio/credentials/config.json.sample')
        # is_expired should be True before retrieve
        eq_(provider.is_expired(), True)
        # retrieve credentials
        creds = provider.retrieve()
        expected_creds = Value(
            access_key='accessKey',
            secret_key='secret',
            session_token=None
        )
        eq_(creds.access_key, expected_creds.access_key)
        eq_(creds.secret_key, expected_creds.secret_key)
        eq_(creds.session_token, expected_creds.session_token)
        # is_expired should be False after retrieve
        eq_(provider.is_expired(), False)

    def test_file_minio_env_alias(self):
        # clear environment
        os.environ.clear()
        # set env with minio config file
        os.environ['MINIO_ALIAS'] = 'play'
        # get provider
        provider = FileMinioClient('minio/credentials/config.json.sample')
        # is_expired should be True before retrieve
        eq_(provider.is_expired(), True)
        # retrieve credentials
        creds = provider.retrieve()
        expected_creds = Value(
            access_key='Q3AM3UQ867SPQQA43P2F',
            secret_key='zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG',
            session_token=None
        )
        eq_(creds.access_key, expected_creds.access_key)
        eq_(creds.secret_key, expected_creds.secret_key)
        eq_(creds.session_token, expected_creds.session_token)
        # is_expired is False after retrieve
        eq_(provider.is_expired(), False)

    def test_file_minio_arg_alias(self):
        # clear environment
        os.environ.clear()
        # get provider
        provider = FileMinioClient(
            'minio/credentials/config.json.sample', 'play')
        # is_expired is True before retrieve
        eq_(provider.is_expired(), True)
        # get credentials
        creds = provider.retrieve()
        expected_creds = Value(
            access_key='Q3AM3UQ867SPQQA43P2F',
            secret_key='zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG',
            session_token=None
        )
        eq_(creds.access_key, expected_creds.access_key)
        eq_(creds.secret_key, expected_creds.secret_key)
        eq_(creds.session_token, expected_creds.session_token)
        # is_expired should be False after retrieve
        eq_(provider.is_expired(), False)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from minio import Minio
from minio.error import ResponseError
from minio.select.errors import SelectCRCValidationError, SelectMessageError
from minio.select.options import (SelectObjectOptions, RequestProgress,
                                  InputSerialization, OutputSerialization)
from minio.select.options import CSVInput
from minio.select.options import CSVOutput
# from minio.select.options import JSONOutput
# from minio.select.options import JsonInput
# from minio.select.options import ParquetInput

client = Minio('s3.amazonaws.com',
               access_key='YOUR-ACCESSKEY',
               secret_key='YOUR-SECRETKEY')

options = SelectObjectOptions(
    expression="select * from s3object",
    input_serialization=InputSerialization(
        compression_type="NONE",
        csv=CSVInput(
            FileHeaderInfo="USE",
            RecordDelimiter="\n",
            FieldDelimiter=",",
            QuoteCharacter='"',
            QuoteEscapeCharacter='"',
            Comments="#",
            AllowQuotedRecordDelimiter="FALSE",
        ),
        # If input is JSON
        # json=JSONInput(Type="DOCUMENT")
    ),

    output_serialization=OutputSerialization(
        csv=CSVOutput(
            QuoteFields="ASNEEDED",
            RecordDelimiter="\n",
            FieldDelimiter=",",
            QuoteCharacter='"',
            QuoteEscapeCharacter='"',
        ),

        # json = JsonOutput(RecordDelimiter="\n")
    ),
    request_progress=RequestProgress(
        enabled="False"
    )
)

try:
    data = client.select_object_content('your-bucket', 'your-object', options)

    # Get the records
    with open('my-record-file', 'w') as record_data:
        for d in data.stream(10*1024):
            record_data.write(d)

    # Get the stats
    print(data.stats())

except SelectMessageError as err:
    print(err)

except SelectCRCValidationError as err:
    print(err)

except ResponseError as err:
    print(err)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from minio.credentials.credentials import Credentials, Value
from minio.credentials.file_minio_client import FileMinioClient
from unittest import TestCase
from nose.tools import eq_


class CredentialsTest(TestCase):
    def test_credentials_get(self):
        # get credentials
        credentials = Credentials(
            provider=FileMinioClient(
                'minio/credentials/config.json.sample', 'play')
        )
        # is_expired should be True before get
        eq_(credentials.is_expired(), True)
        creds = credentials.get()
        expected_creds = Value(
            access_key='Q3AM3UQ867SPQQA43P2F',
            secret_key='zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG',
            session_token=None
        )
        eq_(creds.access_key, expected_creds.access_key)
        eq_(creds.secret_key, expected_creds.secret_key)
        eq_(creds.session_token, expected_creds.session_token)
        # is_expired should be False after get
        eq_(credentials.is_expired(), False)

    def test_credentials_with_error(self):
        pass



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2015, 2016, 2017, 2018, 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.xml_marshal
~~~~~~~~~~~~~~~

This module contains the simple wrappers for XML marshaller's.

:copyright: (c) 2015 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from __future__ import absolute_import
import io

import xml.etree.ElementTree as ET
from collections import defaultdict

_S3_NAMESPACE = 'http://s3.amazonaws.com/doc/2006-03-01/'


def Element(tag, with_namespace=False):
    if with_namespace:
        return ET.Element(tag, {'xmlns': _S3_NAMESPACE})
    return ET.Element(tag)


def SubElement(parent, tag, text=None):
    subElement = ET.SubElement(parent, tag)
    if text is not None:
        subElement.text = text
    return subElement


def get_xml_data(element):
    data = io.BytesIO()
    ET.ElementTree(element).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()


def xml_to_dict(in_xml):
    # Converts xml to dict
    elem = ET.XML(in_xml)
    return etree_to_dict(elem)


def etree_to_dict(elem):
    # Converts ElementTree object to dict
    ns = '{' + _S3_NAMESPACE + '}'
    elem.tag = elem.tag.replace(ns, '')

    d = {elem.tag: {} if elem.attrib else None}
    children = list(elem)
    if children:
        dd = defaultdict(list)
        if children[0].tag.replace(ns, '') == 'Rule':
            for dc in map(etree_to_dict, children):
                for k, v in dc.items():
                    dd[k].append([v])
        else:
            for dc in map(etree_to_dict, children):
                for k, v in dc.items():
                    dd[k].append(v)
        d = {elem.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if elem.attrib:
        d[elem.tag].update(('@' + k, v) for k, v in elem.attrib.items())
    if elem.text:
        text = elem.text.strip()
        if children or elem.attrib:
            if text:
                d[elem.tag]['#text'] = text
        else:
            d[elem.tag] = text
    return d


def xml_marshal_bucket_encryption(rules):
    root = Element('ServerSideEncryptionConfiguration')

    if rules:
        # As server supports only one rule, the first rule is taken due to
        # no validation is done at server side.
        apply_element = SubElement(SubElement(root, 'Rule'),
                                   'ApplyServerSideEncryptionByDefault')
        SubElement(apply_element, 'SSEAlgorithm',
                   rules[0]['ApplyServerSideEncryptionByDefault'].get(
                       'SSEAlgorithm', 'AES256'))
        kms_text = rules[0]['ApplyServerSideEncryptionByDefault'].get(
            'KMSMasterKeyID')
        if kms_text:
            SubElement(apply_element, 'KMSMasterKeyID', kms_text)

    return get_xml_data(root)


def xml_marshal_bucket_constraint(region):
    """
    Marshal's bucket constraint based on *region*.

    :param region: Region name of a given bucket.
    :return: Marshalled XML data.
    """
    root = Element('CreateBucketConfiguration', with_namespace=True)
    SubElement(root, 'LocationConstraint', region)
    return get_xml_data(root)


def xml_marshal_select(opts):
    root = Element('SelectObjectContentRequest')
    SubElement(root, 'Expression', opts.expression)
    SubElement(root, 'ExpressionType', 'SQL')

    input_serialization = SubElement(root, 'InputSerialization')
    SubElement(input_serialization, 'CompressionType',
               opts.in_ser.compression_type)

    if opts.in_ser.csv_input:
        csv = SubElement(input_serialization, 'CSV')
        SubElement(csv, 'FileHeaderInfo', opts.in_ser.csv_input.FileHeaderInfo)
        SubElement(csv, 'RecordDelimiter',
                   opts.in_ser.csv_input.RecordDelimiter)
        SubElement(csv, 'FieldDelimiter', opts.in_ser.csv_input.FieldDelimiter)
        SubElement(csv, 'QuoteCharacter', opts.in_ser.csv_input.QuoteCharacter)
        SubElement(csv, 'QuoteEscapeCharacter',
                   opts.in_ser.csv_input.QuoteEscapeCharacter)
        SubElement(csv, 'Comments', opts.in_ser.csv_input.Comments)
        SubElement(csv, 'AllowQuotedRecordDelimiter',
                   opts.in_ser.csv_input.AllowQuotedRecordDelimiter.lower())

    if opts.in_ser.json_input:
        SubElement(SubElement(input_serialization, 'JSON'), 'Type',
                   opts.in_ser.json_input.Type)

    if opts.in_ser.parquet_input:
        SubElement(input_serialization, 'Parquet')

    output_serialization = SubElement(root, 'OutputSerialization')
    if opts.out_ser.csv_output:
        csv = SubElement(output_serialization, 'CSV')
        SubElement(csv, 'QuoteFields', opts.out_ser.csv_output.QuoteFields)
        SubElement(csv, 'RecordDelimiter',
                   opts.out_ser.csv_output.RecordDelimiter)
        SubElement(csv, 'FieldDelimiter',
                   opts.out_ser.csv_output.FieldDelimiter)
        SubElement(csv, 'QuoteCharacter',
                   opts.out_ser.csv_output.QuoteCharacter)
        SubElement(csv, 'QuoteEscapeCharacter',
                   opts.out_ser.csv_output.QuoteEscapeCharacter)

    if opts.out_ser.json_output:
        SubElement(SubElement(output_serialization, 'JSON'), 'RecordDelimiter',
                   opts.out_ser.json_output.RecordDelimiter)

    SubElement(SubElement(root, 'RequestProgress'), 'Enabled',
               opts.req_progress.enabled.lower())

    return get_xml_data(root)


def xml_marshal_complete_multipart_upload(uploaded_parts):
    """
    Marshal's complete multipart upload request based on *uploaded_parts*.

    :param uploaded_parts: List of all uploaded parts, ordered by part number.
    :return: Marshalled XML data.
    """
    root = Element('CompleteMultipartUpload', with_namespace=True)
    for uploaded_part in uploaded_parts:
        part = SubElement(root, 'Part')
        SubElement(part, 'PartNumber', str(uploaded_part.part_number))
        SubElement(part, 'ETag', '"' + uploaded_part.etag + '"')

    return get_xml_data(root)


def xml_marshal_bucket_notifications(notifications):
    """
    Marshals the notifications structure for sending to S3 compatible storage

    :param notifications: Dictionary with following structure:

    {
        'TopicConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|'s3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|'s3:ObjectCreated:Copy'|'s3:ObjectCreated:CompleteMultipartUpload'|'s3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|'s3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ],
        'QueueConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|'s3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|'s3:ObjectCreated:Copy'|'s3:ObjectCreated:CompleteMultipartUpload'|'s3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|'s3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ],
        'CloudFunctionConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|'s3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|'s3:ObjectCreated:Copy'|'s3:ObjectCreated:CompleteMultipartUpload'|'s3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|'s3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ]
    }

    :return: Marshalled XML data
    """
    root = Element('NotificationConfiguration', with_namespace=True)
    _add_notification_config_to_xml(
        root,
        'TopicConfiguration',
        notifications.get('TopicConfigurations', [])
    )
    _add_notification_config_to_xml(
        root,
        'QueueConfiguration',
        notifications.get('QueueConfigurations', [])
    )
    _add_notification_config_to_xml(
        root,
        'CloudFunctionConfiguration',
        notifications.get('CloudFunctionConfigurations', [])
    )

    return get_xml_data(root)


NOTIFICATIONS_ARN_FIELDNAME_MAP = {
    'TopicConfiguration': 'Topic',
    'QueueConfiguration': 'Queue',
    'CloudFunctionConfiguration': 'CloudFunction',
}


def _add_notification_config_to_xml(node, element_name, configs):
    """
    Internal function that builds the XML sub-structure for a given
    kind of notification configuration.

    """
    for config in configs:
        config_node = SubElement(node, element_name)

        if 'Id' in config:
            SubElement(config_node, 'Id', config['Id'])

        SubElement(config_node, NOTIFICATIONS_ARN_FIELDNAME_MAP[element_name],
                   config['Arn'])

        for event in config['Events']:
            SubElement(config_node, 'Event', event)

        filter_rules = config.get('Filter', {}).get(
            'Key', {}).get('FilterRules', [])
        if filter_rules:
            s3key_node = SubElement(SubElement(config_node, 'Filter'), 'S3Key')
            for filter_rule in filter_rules:
                filter_rule_node = SubElement(s3key_node, 'FilterRule')
                SubElement(filter_rule_node, 'Name', filter_rule['Name'])
                SubElement(filter_rule_node, 'Value', filter_rule['Value'])
    return node


def xml_marshal_delete_objects(object_names):
    """
    Marshal Multi-Object Delete request body from object names.

    :param object_names: List of object keys to be deleted.
    :return: Serialized XML string for multi-object delete request body.
    """
    root = Element('Delete')

    # use quiet mode in the request - this causes the S3 Server to
    # limit its response to only object keys that had errors during
    # the delete operation.
    SubElement(root, 'Quiet', "true")

    # add each object to the request.
    for object_name in object_names:
        SubElement(SubElement(root, 'Object'), 'Key', object_name)

    return get_xml_data(root)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .credentials import Provider, Value


class Static(Provider):
    def __init__(self, access_key=None, secret_key=None, token=None):
        super(Static, self).__init__()
        self._access_key = access_key
        self._secret_key = secret_key
        self._session_token = token

    def retrieve(self):
        return Value(
            access_key=self._access_key,
            secret_key=self._secret_key,
            session_token=self._session_token
        )

    def is_expired(self):
        return False



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import configparser

from .credentials import Provider, Value


class FileAWSCredentials(Provider):
    def __init__(self, filename=None, profile=None, retrieved=False):
        super(Provider, self).__init__()
        self._filename = (
            filename or
            os.environ.get('AWS_SHARED_CREDENTIALS_FILE') or
            os.path.join(os.environ.get('HOME'), '.aws', 'credentials')
        )
        self._profile = profile or os.environ.get('AWS_PROFILE') or "default"
        self._retrieved = retrieved

    def retrieve(self):
        self._retrieved = False
        ini_profile = configparser.ConfigParser()
        ini_profile.read(self._filename)
        access_key = secret = session_token = ''
        try:
            access_key = ini_profile.get(self._profile, 'aws_access_key_id')
            secret = ini_profile.get(self._profile, 'aws_secret_access_key')
            session_token = ini_profile.get(self._profile, 'aws_session_token')
        except:
            pass

        if access_key and secret:
            self._retrieved = True
            return Value(
                access_key=access_key,
                secret_key=secret,
                session_token=session_token)

        return Value()

    def is_expired(self):
        return not self._retrieved



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from xml.etree import ElementTree

from .credentials import Value
from ..helpers import _iso8601_to_utc_datetime

_XML_NS = {
    's3': 'http://s3.amazonaws.com/doc/2006-03-01/',
    'sts': 'https://sts.amazonaws.com/doc/2011-06-15/'
}


def parse_iam_credentials(data):
    """
    Parser for IAM Instance Metadata Security Credentials.

    https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    :param data: Dict containing the json response.
    :return: A 2-tuple containing:
        - a :class:`~minio.credentials.Value` instance with the temporary credentials.
        - A :class:`DateTime` instance of when the credentials expire.
    """
    expiration = _iso8601_to_utc_datetime(data['Expiration'])
    return Value(
        access_key=data['AccessKeyId'],
        secret_key=data['SecretAccessKey'],
        session_token=data['Token']
    ), expiration


def parse_assume_role(data):
    """
    Parser for assume role response.

    :param data: XML response data for STS assume role as a string.
    :return: A 2-tuple containing:
        - a :class:`~minio.credentials.Value` instance with the temporary credentials.
        - A :class:`DateTime` instance of when the credentials expire.
    """
    root = ElementTree.fromstring(data)
    credentials_elem = root.find("sts:AssumeRoleResult", _XML_NS).find(
        "sts:Credentials", _XML_NS)

    access_key = credentials_elem.find("sts:AccessKeyId", _XML_NS).text
    secret_key = credentials_elem.find("sts:SecretAccessKey", _XML_NS).text
    session_token = credentials_elem.find("sts:SessionToken", _XML_NS).text

    expiry_str = credentials_elem.find("sts:Expiration", _XML_NS).text
    expiry = _iso8601_to_utc_datetime(expiry_str)

    return Value(access_key, secret_key, session_token), expiry



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2017 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.fold_case_dict

This module implements a case insensitive dictionary.

:copyright: (c) 2017 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""


class FoldCaseDict(dict):
    def __init__(self, dictionary={}):
        self._data = self.__create(dictionary)

    def __create(self, value):
        if not isinstance(value, dict):
            return value

        data = {}
        for k, v in value.items():
            data[k.lower()] = FoldCaseDict(v) if isinstance(v, dict) else v
        return data

    def __getitem__(self, key):
        return self._data.__getitem__(key.lower())

    def __contains__(self, key):
        return self._data.__contains__(key.lower())

    def __setitem__(self, key, value):
        self._data.__setitem__(key.lower(), self.__create(value))

    def __delitem__(self, key):
        self._data.__delitem__(key.lower())

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return self._data.__len__()

    def __eq__(self, other):
        if isinstance(other, dict):
            other = FoldCaseDict(other)
        elif isinstance(other, FoldCaseDict):
            pass
        else:
            raise NotImplementedError

        return self._data.__eq__(other._data)

    def __repr__(self):
        return self._data.__repr__()

    def get(self, key, default=None):
        return self._data.get(key.lower(), default)

    def has_key(self, key):
        return self._data.has_key(key.lower())

    def items(self):
        return self._data.items()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def iteritems(self):
        return self._data.iteritems()

    def iterkeys(self):
        for k in self._data.keys():
            yield k

    def itervalues(self):
        for v in self._data.values():
            yield v

    def update(self, dictionary):
        if isinstance(dictionary, dict):
            dictionary = FoldCaseDict(dictionary)
        elif isinstance(dictionary, FoldCaseDict):
            pass
        else:
            raise TypeError

        self._data.update(dictionary._data)

    def copy(self):
        return FoldCaseDict(self._data.copy())

    def clear(self):
        self._data = {}

    def pop(self, key):
        return self._data.pop(key.lower())

    def popitem(self):
        return self._data.popitem()



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2016 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: YOUR-ACCESSKEYID, YOUR-SECRETACCESSKEY, my-testfile, my-bucketname and
# my-objectname are dummy values, please replace them with original values.

from datetime import datetime, timezone

from minio import Minio, CopyConditions
from minio.error import ResponseError

client = Minio('s3.amazonaws.com',
               access_key='YOUR-ACCESSKEY',
               secret_key='YOUR-SECRETKEY')

# client.trace_on(sys.stderr)
copy_conditions = CopyConditions()
# Set modified condition, copy object modified since 1st April 2014.
mod_since = datetime(2014, 4, 1, tzinfo=timezone.utc)
copy_conditions.set_modified_since(mod_since)

# Set unmodified condition, copy object unmodified since 1st April 2014.
# copy_conditions.set_unmodified_since(mod_since)

# Set matching ETag condition, copy object which matches the following ETag.
# copy_conditions.set_match_etag("31624deb84149d2f8ef9c385918b653a")

# Set matching ETag except condition, copy object which does not match the
# following ETag.
# copy_conditions.set_match_etag_except("31624deb84149d2f8ef9c385918b653a")

try:
    copy_result = client.copy_object("my-bucket", "my-object",
                                     "/my-sourcebucket/my-sourceobject",
                                     copy_conditions)
    print(copy_result)
except ResponseError as err:
    print(err)



from .static import Static
from .credentials import Credentials, Value
from .chain import Chain
from .assume_role import AssumeRoleProvider
from .aws_iam import IamEc2MetaData
from .env_aws import EnvAWS
from .env_minio import EnvMinio



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015-2020 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
minio.signer
~~~~~~~~~~~~~~~

This module implements all helpers for AWS Signature version '4' support.

:copyright: (c) 2015 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

import collections
import hashlib
import hmac

from datetime import datetime
from .error import InvalidArgumentError
from .compat import urlsplit, queryencode
from .helpers import get_sha256_hexdigest
from .fold_case_dict import FoldCaseDict

# Signature version '4' algorithm.
_SIGN_V4_ALGORITHM = 'AWS4-HMAC-SHA256'

# Hardcoded S3 header value for X-Amz-Content-Sha256
_UNSIGNED_PAYLOAD = u'UNSIGNED-PAYLOAD'

# Hardcoded service name for pre-signed S3 urls
_PRESIGNED_SERVICE_NAME = "s3"

# Default service name for all signature
_DEFAULT_SERVICE_NAME = "s3"


def post_presign_signature(date, region, secret_key, policy_str):
    """
    Calculates signature version '4' for POST policy string.

    :param date: datetime formatted date.
    :param region: region of the bucket for the policy.
    :param secret_key: Amazon S3 secret access key.
    :param policy_str: policy string.
    :return: hexlified sha256 signature digest.
    """
    signing_key = generate_signing_key(date, region, secret_key)
    signature = hmac.new(signing_key, policy_str.encode('utf-8'),
                         hashlib.sha256).hexdigest()

    return signature


def presign_v4(method, url, credentials,
               region=None, headers=None, expires=None,
               response_headers=None, request_date=None):
    """
    Calculates signature version '4' for regular presigned URLs.

    :param method: Method to be presigned examples 'PUT', 'GET'.
    :param url: URL to be presigned.
    :param credentials: Credentials object with your AWS s3 account info.
    :param region: region of the bucket, it is optional.
    :param headers: any additional HTTP request headers to
       be presigned, it is optional.
    :param expires: final expiration of the generated URL. Maximum is 7days.
    :param response_headers: Specify additional query string parameters.
    :param request_date: the date of the request.
    """

    # Validate input arguments.
    if not credentials.get().access_key or not credentials.get().secret_key:
        raise InvalidArgumentError('Invalid access_key and secret_key.')

    region = region or 'us-east-1'
    headers = headers or {}
    expires = expires or '604800'
    request_date = request_date or datetime.utcnow()

    # If a sha256sum is known, add to headers to include with signature
    content_hash_hex = _UNSIGNED_PAYLOAD
    for k in headers:
        if k.lower() == 'x-amz-content-sha256':
            content_hash_hex = headers[k] or _UNSIGNED_PAYLOAD
            del headers[k]
            break

    parsed_url = urlsplit(url)
    host = remove_default_port(parsed_url)
    headers['Host'] = host
    iso8601Date = request_date.strftime("%Y%m%dT%H%M%SZ")

    headers_to_sign = headers
    # Construct queries.
    query = {}
    query['X-Amz-Algorithm'] = _SIGN_V4_ALGORITHM
    query['X-Amz-Credential'] = generate_credential_string(
        credentials.get().access_key, request_date, region, _PRESIGNED_SERVICE_NAME)
    query['X-Amz-Date'] = iso8601Date
    query['X-Amz-Expires'] = str(expires)
    if credentials.get().session_token:
        query['X-Amz-Security-Token'] = credentials.get().session_token

    signed_headers = get_signed_headers(headers_to_sign)
    query['X-Amz-SignedHeaders'] = ';'.join(signed_headers)

    if response_headers:
        query.update(response_headers)

    # URL components.
    url_components = [parsed_url.geturl()]
    ordered_query = collections.OrderedDict(sorted(query.items()))
    query_components = []
    for component_key in ordered_query:
        single_component = [component_key, '=']
        if ordered_query[component_key]:
            single_component.append(queryencode(ordered_query[component_key]))
        query_components.append(''.join(single_component))

    query_string = '&'.join(query_components)
    if query_string:
        url_components.append('?')
        url_components.append(query_string)
    new_url = ''.join(url_components)
    # new url constructor block ends.
    new_parsed_url = urlsplit(new_url)

    canonical_request = generate_canonical_request(method,
                                                   new_parsed_url,
                                                   headers_to_sign,
                                                   signed_headers,
                                                   content_hash_hex)
    string_to_sign = generate_string_to_sign(request_date, region,
                                             canonical_request,
                                             _PRESIGNED_SERVICE_NAME)
    signing_key = generate_signing_key(request_date, region,
                                       credentials.get().secret_key)
    signature = hmac.new(signing_key, string_to_sign.encode('utf-8'),
                         hashlib.sha256).hexdigest()
    new_parsed_url = urlsplit(new_url + "&X-Amz-Signature="+signature)
    return new_parsed_url.geturl()


def get_signed_headers(headers):
    """
    Get signed headers.

    :param headers: input dictionary to be sorted.
    """
    return sorted([h.lower().strip() for h in headers])


def sign_v4(method, url, region, headers=None,
            credentials=None,
            content_sha256=None,
            request_datetime=None,
            service_name=_DEFAULT_SERVICE_NAME
            ):
    """
    Signature version 4.

    :param method: HTTP method used for signature.
    :param url: Final url which needs to be signed.
    :param region: Region should be set to bucket region.
    :param headers: Optional headers for the method.
    :param credentials: Optional Credentials object with your AWS s3 account
                        info.
    :param content_sha256: Optional body sha256.
    :param request_datetime: Optional request date/time
    :param service_name: Optional service to sign request for (defaults to S3)
    """

    # If no access key or secret key is provided return headers.
    if not credentials.get().access_key or not credentials.get().secret_key:
        return headers

    headers = headers or FoldCaseDict()
    region = region or 'us-east-1'

    parsed_url = urlsplit(url)
    secure = parsed_url.scheme == 'https'
    if secure and not content_sha256:
        content_sha256 = _UNSIGNED_PAYLOAD
    content_sha256 = content_sha256 or get_sha256_hexdigest('')

    host = remove_default_port(parsed_url)
    headers['Host'] = host

    request_datetime = request_datetime or datetime.utcnow()

    headers['X-Amz-Date'] = request_datetime.strftime("%Y%m%dT%H%M%SZ")
    headers['X-Amz-Content-Sha256'] = content_sha256
    if credentials.get().session_token:
        headers['X-Amz-Security-Token'] = credentials.get().session_token

    headers_to_sign = headers

    signed_headers = get_signed_headers(headers_to_sign)
    canonical_req = generate_canonical_request(method,
                                               parsed_url,
                                               headers_to_sign,
                                               signed_headers,
                                               content_sha256
                                               )

    string_to_sign = generate_string_to_sign(request_datetime, region,
                                             canonical_req, service_name)
    signing_key = generate_signing_key(
        request_datetime, region, credentials.get().secret_key, service_name)
    signature = hmac.new(signing_key, string_to_sign.encode('utf-8'),
                         hashlib.sha256).hexdigest()

    authorization_header = generate_authorization_header(
        credentials.get().access_key, request_datetime, region, signed_headers,
        signature, service_name)

    headers['Authorization'] = authorization_header
    return headers


def generate_canonical_request(method, parsed_url, headers, signed_headers,
                               content_sha256):
    """
    Generate canonical request.

    :param method: HTTP method.
    :param parsed_url: Parsed url is input from :func:`urlsplit`
    :param headers: HTTP header dictionary.
    :param content_sha256: Content sha256 hexdigest string.
    """
    # Should not encode ~. Decode it back if present.
    parsed_url_path = parsed_url.path.replace("%7E", "~")
    parsed_url_query = parsed_url.query.replace("%7E", "~")
    lines = [method, parsed_url_path, parsed_url_query]

    # Headers added to canonical request.
    header_lines = []
    for header in signed_headers:
        value = headers[header.title()]
        value = str(value).strip()
        header_lines.append(header + ':' + value)

    lines = lines + header_lines
    lines.append('')

    lines.append(';'.join(signed_headers))
    lines.append(content_sha256)
    return '\n'.join(lines)


def generate_string_to_sign(date, region, canonical_request, service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate string to sign.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param canonical_request: Canonical request generated previously.
    :param service_name: Service to scope request for.
    """
    formatted_date_time = date.strftime("%Y%m%dT%H%M%SZ")

    canonical_request_hasher = hashlib.sha256()
    canonical_request_hasher.update(canonical_request.encode('utf-8'))
    canonical_request_sha256 = canonical_request_hasher.hexdigest()
    scope = generate_scope_string(date, region, service_name)

    return '\n'.join([_SIGN_V4_ALGORITHM,
                      formatted_date_time,
                      scope,
                      canonical_request_sha256])


def generate_signing_key(date, region, secret_key, service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate signing key.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param secret_key: Secret access key.
    :param service_name: The signing key is scoped to a service e.g. s3
    """
    formatted_date = date.strftime("%Y%m%d")

    key1_string = 'AWS4' + secret_key
    key1 = key1_string.encode('utf-8')
    key2 = hmac.new(key1, formatted_date.encode('utf-8'),
                    hashlib.sha256).digest()
    key3 = hmac.new(key2, region.encode('utf-8'), hashlib.sha256).digest()
    key4 = hmac.new(key3, service_name.encode(
        'utf-8'), hashlib.sha256).digest()

    return hmac.new(key4, 'aws4_request'.encode('utf-8'),
                    hashlib.sha256).digest()


def generate_scope_string(date, region, service_name):
    """
    Generate scope string.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param service_name: Service for scope string, e.g., "s3".
    """
    formatted_date = date.strftime("%Y%m%d")
    scope = '/'.join([formatted_date,
                      region,
                      service_name,
                      'aws4_request'])
    return scope


def generate_credential_string(access_key, date, region, service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate credential string.

    :param access_key: Server access key.
    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param service_name: Service to scope credentials to.
    """
    return access_key + '/' + generate_scope_string(date, region, service_name)


def generate_authorization_header(access_key, date, region,
                                  signed_headers, signature,
                                  service_name=_DEFAULT_SERVICE_NAME):
    """
    Generate authorization header.

    :param access_key: Server access key.
    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param signed_headers: Signed headers.
    :param signature: Calculated signature.
    :param service_name: Optional service to sign request for.
    """
    signed_headers_string = ';'.join(signed_headers)
    credential = generate_credential_string(
        access_key, date, region, service_name)
    auth_header = [_SIGN_V4_ALGORITHM, 'Credential=' + credential + ',',
                   'SignedHeaders=' + signed_headers_string + ',',
                   'Signature=' + signature]
    return ' '.join(auth_header)


def remove_default_port(parsed_url):
    default_ports = {
        'http': 80,
        'https': 443
    }
    if any(parsed_url.scheme == scheme and parsed_url.port == port
           for scheme, port in default_ports.items()):
        # omit default port (i.e. 80 or 443)
        host = parsed_url.hostname
    else:
        host = parsed_url.netloc
    return host
